#!/usr/bin/env python

__author__ = "Patrick Godwin (patrick.godwin@ligo.org)"
__description__ = "tools to serve data and dynamic html pages"

#-------------------------------------------------
### imports

import copy
import functools
import json
import os
import pkg_resources
import sys
import time

from six.moves import urllib

import bottle
import yaml

from . import aggregator
from . import io
from . import templates
from . import utils

#-------------------------------------------------
### constants + env variables

JSON_HEADER = {
    'Content-type': 'application/json',
    'Cache-Control': 'max-age=10',
}


#-------------------------------------------------
### bottle apps/routing

def static(file_):
    static_dir = pkg_resources.resource_filename(pkg_resources.Requirement.parse('ligo_scald'), 'static')
    yield bottle.static_file(file_, root=static_dir)


def index(config, script_name='', use_cgi=False):
    config = copy.deepcopy(config)
    static_dir = '../' if use_cgi else ''

    # check if the user has checked any plots, otherwise use default values
    if any(["livecharts" in k for k in bottle.request.query]):
        for plot in config['plots']:
            plot['value']='checked' if "livecharts%s" % plot['title'].replace(' ', '_').lower() in bottle.request.query else ''

    if "GPS" in bottle.request.query:
        config['dashboard']['GPS']=bottle.request.query['GPS']
    if "Duration" in bottle.request.query:
        config['dashboard']['Duration']=bottle.request.query['Duration']

    ### format query params for plots
    for plot in config['plots']:
        if 'params' in plot:
            plot['params'] = urllib.parse.urlencode([tuple(param) for param in plot['params']])
        else:
            plot['params'] = ''

    ### generate html
    yield bottle.template(templates.HEADER_TEMPLATE, title=config['dashboard']['title'], static_dir=static_dir)

    yield bottle.template(templates.BODY_HEADER_TEMPLATE)
    yield bottle.template(
        templates.DASHBOARD_NAVBAR_TEMPLATE,
        title=config['dashboard']['title'],
        image=config['dashboard']['image'],
        tabs=config['dashboard']['tabs'],
        plots=config['plots'],
        gps=config['dashboard']['GPS'],
        duration=config['dashboard']['Duration'],
        static_dir=static_dir,
    )


    ### find out whether query is for realtime or historical data
    ### NOTE: GPS < 0 is used as a proxy for realtime data and refers
    ###       to the delay (i.e. -5 means to query up to now - 5s
    gpstime = int(config['dashboard']['GPS'])
    if gpstime < 0:
        delay = -gpstime
        stop = aggregator.now() - delay
        refresh = 2000
    else:
        delay = 0
        stop = gpstime
        refresh = -1
    start = stop - int(config['dashboard']['Duration'])

    yield bottle.template(
        templates.DASHBOARD_PLOT_TEMPLATE,
        plots=config['plots'],
        start=start,
        stop=stop,
        refresh=refresh,
        delay=delay,
        script_name=script_name,
    )

    yield bottle.template(templates.BODY_FOOTER_TEMPLATE)


def serve_raw_data(measurement, column, start, end, config):
    ### parse query
    if 'datetime' in bottle.request.query:
        use_datetime = bottle.request.query['datetime'] == 'true'
    else:
        use_datetime = False

    tags = [(key, val) for key, val in bottle.request.query.allitems() if key != 'datetime']

    response = []
    backend = config['database']['backend']
    if backend == 'influxdb':
        ### set up HTTP client
        db_name = config['database']['name']
        hostname = config['database']['hostname']
        port = config['database']['port'] if 'port' in config['database'] else 8086
        client = io.influx.create_client(host=hostname, port=port)

        ### query db for timeseries
        if tags:
            for tag in tags:
                time, data = io.influx.retrieve_timeseries(client, db_name, measurement, start, end, column, tags=[tag], datetime=use_datetime)
                response.append({'x':time, 'y':data, 'name': tag[1]})
        else:
            time, data = io.influx.retrieve_timeseries(client, db_name, measurement, start, end, column, tags=tags, datetime=use_datetime)
            response.append({'x':time, 'y':data})

    elif backend == 'hdf5':
        rootdir = config['database']['rootdir']
        basename = config['database']['basename'] if 'basename' in config['database'] else 'aggregator'

        ### query file system for timeseries
        if tags:
            for tag in tags:
                time, data = io.hdf5.retrieve_timeseries(rootdir, basename, measurement, start, end, column, tags=[tag])
                response.append({'x':time, 'y':data, 'name': tag[1]})
        else:
            time, data = io.hdf5.retrieve_timeseries(rootdir, basename, measurement, start, end, column, tags=tags)
            response.append({'x':time, 'y':data})

    else:
        raise NotImplementedError

    ### format request
    response = [{'x':time, 'y':data}]

    ### return data
    return bottle.HTTPResponse(status=200, headers=JSON_HEADER, body=json.dumps(response).replace("NaN","null"))


def serve_data(measurement, column, start, end, aggregate, dt, config):
    ### parse query
    if 'datetime' in bottle.request.query:
        use_datetime = bottle.request.query['datetime'] == 'true'
    else:
        use_datetime = False

    tags = [(key, val) for key, val in bottle.request.query.allitems() if key != 'datetime']

    response = []
    backend = config['database']['backend']
    if backend == 'influxdb':

        ### set up HTTP client
        db_name = config['database']['name']
        hostname = config['database']['hostname']
        port = config['database']['port'] if 'port' in config['database'] else 8086
        client = io.influx.create_client(host=hostname, port=port)

        ### query db for timeseries
        if tags:
            for tag in tags:
                time, data = io.influx.retrieve_timeseries(client, db_name, measurement, start, end, column, tags=[tag], aggregate=aggregate, dt=dt, datetime=use_datetime)
                response.append({'x':time, 'y':data, 'name': tag[1]})
        else:
            time, data = io.influx.retrieve_timeseries(client, db_name, measurement, start, end, column, tags=tags, aggregate=aggregate, dt=dt, datetime=use_datetime)
            response.append({'x':time, 'y':data})

    elif backend == 'hdf5':
        rootdir = config['database']['rootdir']
        basename = config['database']['basename'] if 'basename' in config['database'] else 'aggregator'

        ### query file system for timeseries
        if tags:
            for tag in tags:
                time, data = io.hdf5.retrieve_timeseries(rootdir, basename, measurement, start, end, column, tags=[tag], aggregate=aggregate, dt=dt)
                response.append({'x':time, 'y':data, 'name': tag[1]})
        else:
            time, data = io.hdf5.retrieve_timeseries(rootdir, basename, measurement, start, end, column, tags=tags, aggregate=aggregate, dt=dt)
            response.append({'x':time, 'y':data})

    ### return data
    return bottle.HTTPResponse(status=200, headers=JSON_HEADER, body=json.dumps(response).replace("NaN","null"))


def serve_raw_data_by_tag(measurement, column, start, end, tag, config):
    ### parse query
    if 'ifo' in bottle.request.query:
        ifos = bottle.request.query['ifo']
    else:
        ifos = None

    if 'tag' in bottle.request.query:
        tags = bottle.request.query['tag']
    else:
        tags = []

    if 'datetime' in bottle.request.query:
        use_datetime = bottle.request.query['datetime'] == 'true'
    else:
        use_datetime = False

    backend = config['database']['backend']
    if backend == 'influxdb':
        ### set up HTTP client
        db_name = config['database']['name']
        hostname = config['database']['hostname']
        port = config['database']['port'] if 'port' in config['database'] else 8086
        client = io.influx.create_client(host=hostname, port=port)

        ### query db for timeseries
        timearray, tagarray, dataarray = io.influx.retrieve_binnedtimeseries_by_tag(client, db_name, measurement, start, end, column, tag, datetime=use_datetime)

    elif backend == 'hdf5':
        rootdir = config['database']['rootdir']
        basename = config['database']['basename'] if 'basename' in config['database'] else 'aggregator'

        ### query file system for timeseries
        timearray, tagarray, dataarray = io.hdf5.retrieve_binnedtimeseries_by_tag(rootdir, basename, measurement, start, end, column, tag)

    ### format request
    response = [{'x':timearray, 'y':tagarray, 'z':dataarray}]

    ### return data
    return bottle.HTTPResponse(status=200, headers=JSON_HEADER, body=json.dumps(response).replace("NaN","null"))


def serve_data_by_tag(measurement, column, start, end, aggregate, dt, tag, config):
    ### parse query
    if 'ifo' in bottle.request.query:
        ifos = bottle.request.query['ifo']
    else:
        ifos = None

    if 'tag' in bottle.request.query:
        tags = bottle.request.query['tag']
    else:
        tags = []

    if 'datetime' in bottle.request.query:
        use_datetime = bottle.request.query['datetime'] == 'true'
    else:
        use_datetime = False

    backend = config['database']['backend']
    if backend == 'influxdb':
        ### set up HTTP client
        db_name = config['database']['name']
        hostname = config['database']['hostname']
        port = config['database']['port'] if 'port' in config['database'] else 8086
        client = io.influx.create_client(host=hostname, port=port)

        ### query db for timeseries
        timearray, tagarray, dataarray = io.influx.retrieve_binnedtimeseries_by_tag(client, db_name, measurement, start, end, column, tag, aggregate=aggregate, dt=dt, datetime=use_datetime)

    elif backend == 'hdf5':
        rootdir = config['database']['rootdir']
        basename = config['database']['basename'] if 'basename' in config['database'] else 'aggregator'

        ### query file system for timeseries
        timearray, tagarray, dataarray = io.hdf5.retrieve_binnedtimeseries_by_tag(rootdir, basename, measurement, start, end, column, tag, aggregate=aggregate, dt=dt)

    ### format request
    response = [{'x':timearray, 'y':tagarray, 'z':dataarray}]

    ### return data
    return bottle.HTTPResponse(status=200, headers=JSON_HEADER, body=json.dumps(response).replace("NaN","null"))


def serve_latest_data_by_tag(measurement, column, start, end, aggregate, dt, tag, config):
    ### parse query
    if 'ifo' in bottle.request.query:
        ifos = bottle.request.query['ifo']
    else:
        ifos = None

    if 'tag' in bottle.request.query:
        tags = bottle.request.query['tag']
    else:
        tags = []

    if 'datetime' in bottle.request.query:
        use_datetime = bottle.request.query['datetime'] == 'true'
    else:
        use_datetime = False

    tag = config['measurements'][measurement]['tag']
    tag_ids = [str(tag_num).zfill(4) for tag_num in range(config['measurements'][measurement]['num_tags'])]
    default_value = config['measurements'][measurement]['default']
    transform = config['measurements'][measurement]['transform']

    y = []
    now = utils.gps_now()

    backend = config['database']['backend']
    if backend == 'influxdb':
        ### set up HTTP client
        db_name = config['database']['name']
        hostname = config['database']['hostname']
        port = config['database']['port'] if 'port' in config['database'] else 8086
        client = io.influx.create_client(host=hostname, port=port)

        ### query db for timeseries
        for tag_id in tag_ids:
            time, data = io.influx.retrieve_timeseries_latest(client, db_name, measurement, column, tags=[(tag, tag_id)], aggregate=aggregate, dt=dt, datetime=use_datetime)

            if transform == 'none':
                if data:
                    y.append(data[-1])
                else:
                    y.append(default_value)

            elif transform == 'latency':
                if time:
                    y.append(now - time[-1])
                else:
                    y.append(default_value)

            else:
                raise NotImplementedError('transform option not known/implemented')

    elif backend == 'hdf5':
        rootdir = config['database']['rootdir']
        basename = config['database']['basename'] if 'basename' in config['database'] else 'aggregator'

        # FIXME: overwriting start and dt is a hack to get around missing data, but retrieve functions
        #        should be able to grab the last data point without needing a range
        start = end - 10000
        dt = 10

        ### query file system for timeseries
        for tag_id in tag_ids:
            time, data = io.hdf5.retrieve_timeseries(rootdir, basename, measurement, start, end, column, tags=[(tag, tag_id)], aggregate=aggregate, dt=dt)

            if transform == 'none':
                if data:
                    y.append(data[-1])
                else:
                    y.append(default_value)

            elif transform == 'latency':
                if time:
                    y.append(now - time[-1])
                else:
                    y.append(default_value)

            else:
                raise NotImplementedError('transform option not known/implemented')

    ### format request
    response = [{'x':tag_ids, 'y':y}]

    ### return data
    return bottle.HTTPResponse(status=200, headers=JSON_HEADER, body=json.dumps(response))


def serve_nagios(config):
    # FIXME: only works for a single type of monitor, need to be able to generalize to other processes
    backend = config['database']['backend']

    duration = config['nagios']['lookback']
    end = aggregator.now()
    start = end - duration

    if duration <= 1000:
        dt = 1
    elif duration <= 10000:
        dt = 10
    elif duration <= 100000:
        dt = 100
    elif duration <= 1000000:
        dt = 1000
    elif duration <= 10000000:
        dt = 10000
    else:
        dt = 100000

    measurement = config['nagios']['measurement']
    column = config['nagios']['column']
    aggregate = config['nagios']['aggregate']
    tag = config['nagios']['tag']
    tag_ids = [str(tag_num).zfill(4) for tag_num in range(config['nagios']['num_tags'])]

    bad_status = 0
    delays = []
    now = utils.gps_now()

    if backend == 'influxdb':

        ### set up HTTP client
        db_name = config['database']['name']
        hostname = config['database']['hostname']
        port = config['database']['port'] if 'port' in config['database'] else 8086
        client = io.influx.create_client(host=hostname, port=port)

        ### query db for timeseries
        for tag_id in tag_ids:
            time, _ = io.influx.retrieve_timeseries(client, db_name, measurement, start, end, column, tags=[(tag, tag_id)], aggregate=aggregate, dt=dt)

            if time:
                delays.append(now - time[-1])
            else:
                bad_status += 1

    elif backend == 'hdf5':
        rootdir = config['database']['rootdir']
        basename = config['database']['basename'] if 'basename' in config['database'] else 'aggregator'

        ### query file system for timeseries
        for tag_id in tag_ids:
            time, _ = io.hdf5.retrieve_timeseries(rootdir, basename, measurement, start, end, column, tags=[(tag, tag_id)], aggregate=aggregate, dt=dt)

            if time:
                delays.append(now - time[-1])
            else:
                bad_status += 1

    ### format nagios response
    if bad_status:
        response = {
            "nagios_shib_scraper_ver": 0.1,
            "status_intervals": [{
                "num_status": 2,
                "txt_status": "{} jobs more than 10 mins behind".format(bad_status),
            }],
        }
    else:
        response = {
            "nagios_shib_scraper_ver": 0.1,
            "status_intervals": [{
                "num_status": 0,
                "txt_status": "OK: Max delay: {} s".format(max(delays)),
            }],
        }

    ### return response
    return bottle.HTTPResponse(status=200, headers=JSON_HEADER, body=json.dumps(response, sort_keys=True, indent=4, separators=(',', ': ')))


def serve_sqlite_data(data, start, end, config):
    ### set up HTTP client
    db_name = config['database']['name']
    db_path = config['database']['path']
    table = config['database']['table']
    client = io.sqlite.create_client(db_name, db_path)

    ### query db for timeseries
    time, data = io.sqlite.retrieve_timeseries(client, [data], start, end, table)

    ### format request
    response = [{'x':time, 'y':data}]

    ### return data
    return bottle.HTTPResponse(status=200, headers=JSON_HEADER, body=json.dumps(response).replace("NaN","null"))


#-------------------------------------------------
### functions

def _add_parser_args(parser):
    parser.add_argument('-c', '--config', required=True,
                        help="sets specific plot and dashboard options based on yaml configuration.")
    parser.add_argument('-b', '--backend', default='wsgiref',
                        help="chooses server backend. options: [cgi|wsgiref]. default=wsgiref.")
    parser.add_argument('-e', '--with-cgi-extension', default=False, action='store_true',
                        help="chooses whether scripts need to have a .cgi extension (if using cgi backend)")
    parser.add_argument('-n', '--application-name',
                        help="chooses the web application name. default = yaml configuration name.")


#-------------------------------------------------
### main

def main(args=None):
    """Serves data and dynamic html pages

    """
    if not args:
        parser = argparse.ArgumentParser()
        _parser_add_arguments(parser)
        args = parser.parse_args()

    ### parse args and set up configuration
    server_backend = args.backend
    app_name = args.application_name

    ### hacks to deal with running on apache and/or cgi
    use_cgi = (server_backend == 'cgi')

    if not app_name:
        app_name = os.path.splitext(os.path.basename(args.config))[0]
    if args.with_cgi_extension:
        app_name += '.cgi/'
    else:
        app_name += '/'
    script_name = app_name if use_cgi else ''

    config = None
    with open(args.config, 'r') as f:
        config = yaml.safe_load(f)

    app = bottle.Bottle()

    ### set up routes
    app.route("/", "GET", functools.partial(index, config=config, use_cgi=use_cgi, script_name=script_name))
    app.route("/api/<measurement>/<column>/raw/<start:int>/<end:int>", "GET", functools.partial(serve_raw_data, config=config))
    app.route("/api/<measurement>/<column>/<aggregate>/<dt:int>/<start:int>/<end:int>", "GET", functools.partial(serve_data, config=config))
    app.route("/api/<measurement>/<column>/by/<tag>/raw/<start:int>/<end:int>", "GET", functools.partial(serve_raw_data_by_tag, config=config))
    app.route("/api/<measurement>/<column>/by/<tag>/<aggregate>/<dt:int>/<start:int>/<end:int>", "GET", functools.partial(serve_data_by_tag, config=config))
    app.route("/api/<measurement>/<column>/by/<tag>/latest/<aggregate>/<dt:int>/<start:int>/<end:int>", "GET", functools.partial(serve_latest_data_by_tag, config=config))
    app.route("/api/nagios", "GET", functools.partial(serve_nagios, config=config))

    ### FIXME: this route should be handled more generally by the other ones but is here to purely handle sqlite queries
    app.route("/api/<data>/sqlite/<start:int>/<end:int>", "GET", functools.partial(serve_sqlite_data, config=config))

    ### serve static files directly
    if server_backend == 'wsgiref':
        app.route("/static/<file_>", "GET", static)

    ### start server
    bottle.run(app, server=server_backend, debug=True)
